using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Actions v1.03
//
// Actions for VAMStory

namespace VAMStoryActionPlugin
{
	public partial class VAMStoryAction : MVRScript
    {
	    // Strings! (yes I know you like that)
	    private static string PLUGIN_PATH;
	    private static string BACKUP_PATH = "Saves\\PluginData\\vamstory\\backups";
	    private static string UI_BUNDLE_PATH = "/vamstory_assets.uibundle";
	    private static string SAVE_PREFIX = "Act_";
	    private static string BACKGROUNDS_PATH = "assets/vamstory/textures/backgrounds/";
	    private static string FONTS_PATH = "assets/vamstory/fonts/";
	    private static string SOUNDS_PATH = "assets/vamstory/sounds/";
	    private static string MATERIALS_PATH = "assets/vamstory/materials/materials_choices/";
	    
	    private string lastFileExplorerPath = "Custom";
	    
		// Game Objects, components, assets and assets lists
		protected Camera VAMSCamera;
		protected GameObject VAMSAUIGO;

		protected Transform TRStoryActionsUI;
		protected Transform TRAudio;
		protected Transform TRActionsMainPanel;
		protected Transform TRActionsPanel;
		protected Transform TRActionBtnModel;
		protected Transform TRActionSlider;
		protected Transform TRActionSliderBg;
		protected Transform TRActionSliderFill;
		protected Transform TRHideUIGrp;
		protected Transform TRHideUIBtn;
		protected Transform TRShowUIGrp;
		protected Transform TRShowUIBtn;

		protected Canvas CANStoryActionsUI;
		protected AudioSource ASAudio;
		protected CanvasGroup CGActionsPanel;
		protected GridLayoutGroup GLGActionsPanel;
		protected LayoutElement LEActionsPanel;
		protected CanvasGroup CGActionSlider;
		protected Slider SLActionSlider;
		protected Image IMGActionSliderBg;
		protected Image IMGActionSliderFill;
		protected Button BTNHideUI;
		protected RawImage IMGHideUI;
		protected Button BTNShowUI;
		protected RawImage IMGShowUI;

		protected Texture2D tmpTexture;
		protected List<string> backgroundsChoices = new List<String>();
		protected List<Sprite> backgroundsSprites = new List<Sprite>();
		protected List<string> fontsChoices = new List<String>();
		protected List<Font> fontsFiles = new List<Font>();
		protected List<string> soundChoices = new List<String>();
		protected List<AudioClip> soundClips = new List<AudioClip>();
		protected List<string> materialChoices = new List<String>() {"None"};
		protected List<Material> materialList = new List<Material>();
		
		// Actions
		protected JSONStorableFloat setSliderValue;
		
		// Static Storables
		protected JSONStorableBool GlobalOptEnableSlider;
		protected EventTrigger GlobalOptSliderTrigger;
		protected JSONStorableBool GlobalOptUIRotationDisabled;
		protected JSONStorableBool GlobalOptUIConstraintToY;
		protected JSONStorableBool GlobalOptUIHideToggle;
		protected JSONStorableFloat GlobalOptUISize;
		protected JSONStorableStringChooser GlobalOptUIAlignment;
		protected JSONStorableFloat GlobalOptBtnAreaWidth;
		protected JSONStorableFloat GlobalOptBtnWidth;
		protected JSONStorableFloat GlobalOptBtnHeight;
		protected JSONStorableFloat GlobalOptBtnSpacing;
		protected JSONStorableFloat GlobalOptFontSize;
		protected JSONStorableStringChooser GlobalOptFont;
		protected JSONStorableStringChooser GlobalOptBg;
		protected JSONStorableStringChooser GlobalOptSoundOnClick;
		protected JSONStorableFloat GlobalOptSoundOnClickVolume;
		protected JSONStorableColor GlobalOptBgColor;
		protected JSONStorableFloat GlobalOptBgOpacity;
		protected JSONStorableColor GlobalOptTxtColor;
		protected JSONStorableStringChooser GlobalOptMaterial;
		protected JSONStorableFloat GlobalOptMatDetSSX;
		protected JSONStorableFloat GlobalOptMatDetSSY;
		protected JSONStorableFloat GlobalOptMatSubDetOpacity;
		protected JSONStorableFloat GlobalOptMatRefraction;
		protected JSONStorableColor GlobalOptMatDetColor;

		// The usual help section
		protected JSONStorableString helpText;
		
		// Actions
		protected List<StoryAction> storyActions = new List<StoryAction>();
		
		// Plugin data (due to the fact that the UI is instantiated after init, I need to grab my data and restore manually
		protected JSONClass actionsDatas { get; set; }
		
		// Dynamic UI
		protected List<UIDynamic> globalControlsUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalOptionsUIs = new List<UIDynamic>();
		protected List<UIDynamic> globalColorsOptUIs = new List<UIDynamic>();
		protected List<UIDynamic> helpUIs = new List<UIDynamic>();
		
		protected List<UIDynamic> ignoreLayoutUI = new List<UIDynamic>();
		
		// Global UI components
		protected JSONStorableStringChooser RightUISelector;
		protected List<string> uisChoices = new List<string> {"Actions controls", "Global options", "Colors/Material options", "Help"};
		protected JSONStorableStringChooser ActionSelector;
		protected List<string> actionsChoices;
		protected JSONStorableStringChooser CopySelector;
		protected List<string> copyChoices = new List<String>();
		protected List<string> alignmentChoices = new List<string> {"Left", "Right"};

		protected UIDynamicTextField RightUIMainTitle;
		
		// Assetbundle
		private Request bundleRequest = null;

		// Init and security flags
		private bool isBundleLoaded = false;
		private bool isInit = false;

		public override void Init()
        {
            try
            {
                if (containingAtom.type != "Empty")
                {
                    SuperController.LogError("Please add VAMStory on an Empty Atom");
                    return;
                }

                PLUGIN_PATH = GetPluginPath(this);

				VAMSCamera = SuperController.singleton.centerCameraTarget.targetCamera;
				
				// Backup folder
				FileManagerSecure.CreateDirectory(BACKUP_PATH);
				StartCoroutine("AutoBackup");
                
                // ******* CREATING MAIN TRIGGERS/ACTIONS ********
                // They are here to be displayed on TOP of every other JSONStorables
                JSONStorableAction fakeFuncUseBelow = new JSONStorableAction("- - - - Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncUseBelow);
                
                setSliderValue = new JSONStorableFloat("Set Slider Value", 0.5f, 0f, 1f) {isStorable=false,isRestorable=false};
                setSliderValue.setCallbackFunction += (val) => { OnSetSliderValue(val); };
                RegisterFloat(setSliderValue);
                
                // ******* CREATING MAIN UI *********************************
                // Temporary vars
                HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
                HSVColor hsvcTxtDefault = HSVColorPicker.RGBToHSV(0.2f, 0.2f, 0.2f);
                UIDynamicTextField tmpTextfield;
                UIDynamicPopup tmpPopup;
                UIDynamicSlider tmpSlider;
                UIDynamicColorPicker tmpColor;
                UIDynamicButton tmpButton;
                UIDynamicToggle tmpToggle;
                UIDynamic tmpSpacer;
                
                // ******* TITLE **********************
                createStaticDescriptionText("VAMStory Title","<color=#000><size=35><b>VAM STORY ACTIONS</b></size></color>",true,40,TextAnchor.MiddleRight);
                
				RightUISelector = new JSONStorableStringChooser("Right UI Selector", uisChoices, uisChoices.First(), "Show") {isStorable=false,isRestorable=false};
                RightUISelector.setCallbackFunction += (val) => { OnChangeRightUISelection(); };
                UIDynamicPopup RUISelsp = CreateScrollablePopup(RightUISelector, true);
                RUISelsp.labelWidth = 95f;

                tmpSpacer = CreateSpacer(true);
                setupSpacer(tmpSpacer, 25f);                                
                
                RightUIMainTitle = createStaticDescriptionText("Right UI Main Title","<color=#000><size=35><b>ACTIONS CONTROLS</b></size></color>",true,55,TextAnchor.MiddleRight);
				
                GlobalOptEnableSlider = new JSONStorableBool("Global_Opt_EnableSlider", false);
                tmpToggle = CreateToggle(GlobalOptEnableSlider,true);
                tmpToggle.label = "Enable Slider";
                RegisterBool(GlobalOptEnableSlider);
                GlobalOptEnableSlider.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpToggle);
                
                GlobalOptSliderTrigger = new EventTrigger(this, "Global_Opt_SliderTrigger");
                GlobalOptSliderTrigger.active = true;
                
                tmpButton = CreateButton("Edit slider trigger",true);
                tmpButton.button.onClick.AddListener( () => { GlobalOptSliderTrigger.OpenPanelTransition(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.8f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);
                
				GlobalOptUIRotationDisabled = new JSONStorableBool("Global_Opt_UIRotationDisabled", false);
				tmpToggle = CreateToggle(GlobalOptUIRotationDisabled,true);
				tmpToggle.label = "UI rotation disabled";
				RegisterBool(GlobalOptUIRotationDisabled);
				GlobalOptUIRotationDisabled.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
				globalOptionsUIs.Add((UIDynamic) tmpToggle);
				
                GlobalOptUIConstraintToY = new JSONStorableBool("Global_Opt_UIConstraintToY", false);
                tmpToggle = CreateToggle(GlobalOptUIConstraintToY,true);
				tmpToggle.label = "UI rotation constrained to Y";
				RegisterBool(GlobalOptUIConstraintToY);
				globalOptionsUIs.Add((UIDynamic) tmpToggle);

				GlobalOptUIHideToggle = new JSONStorableBool("Global_Opt_UIHideToggle", false);
				tmpToggle = CreateToggle(GlobalOptUIHideToggle,true);
				tmpToggle.label = "UI hide toggle button";
				RegisterBool(GlobalOptUIHideToggle);
				GlobalOptUIHideToggle.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
				globalOptionsUIs.Add((UIDynamic) tmpToggle);
				
                GlobalOptUISize = new JSONStorableFloat("Global_Opt_UISize", 1f, 0.5f, 2f);
                tmpSlider = CreateSlider(GlobalOptUISize, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "UI Size";
                RegisterFloat(GlobalOptUISize);
                GlobalOptUISize.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptUIAlignment = new JSONStorableStringChooser("Global_Opt_UIAlignment", alignmentChoices, "Left", "UI Alignment");
                RegisterStringChooser(GlobalOptUIAlignment);
                GlobalOptUIAlignment.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptUIAlignment, true);
                tmpPopup.labelWidth = 220f;
                globalOptionsUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptBtnAreaWidth = new JSONStorableFloat("Global_Opt_BtnAreaWidth", 1000f, 1000f, 2000f);
                tmpSlider = CreateSlider(GlobalOptBtnAreaWidth, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Buttons Area Width";
                RegisterFloat(GlobalOptBtnAreaWidth);
                GlobalOptBtnAreaWidth.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBtnWidth = new JSONStorableFloat("Global_Opt_BtnWidth", 326f, 100f, 1000f);
                tmpSlider = CreateSlider(GlobalOptBtnWidth, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Buttons Width";
                RegisterFloat(GlobalOptBtnWidth);
                GlobalOptBtnWidth.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBtnHeight = new JSONStorableFloat("Global_Opt_BtnHeight", 100f, 25f, 300f);
                tmpSlider = CreateSlider(GlobalOptBtnHeight, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Buttons Height";
                RegisterFloat(GlobalOptBtnHeight);
                GlobalOptBtnHeight.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptBtnSpacing = new JSONStorableFloat("Global_Opt_BtnSpacing", 10f, 0f, 500f);
                tmpSlider = CreateSlider(GlobalOptBtnSpacing, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Buttons Spacing";
                RegisterFloat(GlobalOptBtnSpacing);
                GlobalOptBtnSpacing.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptFontSize = new JSONStorableFloat("Global_Opt_FontSize", 40f, 30f, 60f, false);
                tmpSlider = CreateSlider(GlobalOptFontSize, true);
                tmpSlider.valueFormat = "F0";
                tmpSlider.label = "Font Size";
                RegisterFloat(GlobalOptFontSize);
                GlobalOptFontSize.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptFont = new JSONStorableStringChooser("Global_Opt_Font", fontsChoices, "arial", "Font");
                RegisterStringChooser(GlobalOptFont);
                GlobalOptFont.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptFont, true);
                tmpPopup.labelWidth = 120f;
                globalOptionsUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptSoundOnClick = new JSONStorableStringChooser("Global_Opt_SndClick", soundChoices, "click_02", "Sound On Click");
                RegisterStringChooser(GlobalOptSoundOnClick);
                GlobalOptSoundOnClick.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptSoundOnClick, true);
                tmpPopup.labelWidth = 120f;
                globalOptionsUIs.Add((UIDynamic) tmpPopup);

                GlobalOptSoundOnClickVolume = new JSONStorableFloat("Global_Opt_SndClickVol", 1f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptSoundOnClickVolume, true);
                tmpSlider.valueFormat = "F3";
                tmpSlider.label = "On click sound volume";
                RegisterFloat(GlobalOptSoundOnClickVolume);
                globalOptionsUIs.Add((UIDynamic) tmpSlider);
                
                // Backups
                tmpSpacer = CreateSpacer(true);
                globalOptionsUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 25f);                                

                tmpTextfield = createStaticDescriptionText("Backups Title","<color=#000><size=27><b>BACKUPS</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalOptionsUIs.Add((UIDynamic) tmpTextfield);
                
                tmpButton = CreateButton("Restore backup",true);
                tmpButton.button.onClick.AddListener(() => { LoadBackup(); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.5f, 0.7f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);
                
                tmpButton = CreateButton("Save manual backup",true);
                tmpButton.button.onClick.AddListener(() => { SaveBackup(2); } );
                setButtonColor(tmpButton, new Color(0.1f, 0.7f, 0.4f, 1f));
                setButtonTextColor(tmpButton, new Color(1f, 1f, 1f, 1f));
                globalOptionsUIs.Add((UIDynamic) tmpButton);

                // Colors and materials
                GlobalOptBg = new JSONStorableStringChooser("Global_Opt_Bg", backgroundsChoices, "Default", "Style");
                GlobalOptBg.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptBg, true);
                tmpPopup.labelWidth = 120f;
                RegisterStringChooser(GlobalOptBg);
                globalColorsOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptBgColor = new JSONStorableColor("Global_Opt_BgColor", hsvcDefault);
                GlobalOptBgColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptBgColor, true);
                tmpColor.label = "Action background color";
                RegisterColor(GlobalOptBgColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
				
                GlobalOptBgOpacity = new JSONStorableFloat("Global_Opt_BgOpacity", 1f, 0f, 1f, false);
                GlobalOptBgOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpSlider = CreateSlider(GlobalOptBgOpacity, true);
                tmpSlider.label = "Action background Opacity";
                RegisterFloat(GlobalOptBgOpacity);
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptTxtColor = new JSONStorableColor("Global_Opt_TxtColor", hsvcTxtDefault);
                GlobalOptTxtColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptTxtColor, true);
                tmpColor.label = "Action text Color";
                RegisterColor(GlobalOptTxtColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);

                tmpSpacer = CreateSpacer(true);
                globalColorsOptUIs.Add((UIDynamic) tmpSpacer);
                setupSpacer(tmpSpacer, 25f); 
                
                GlobalOptMaterial = new JSONStorableStringChooser("Global_Opt_Material", materialChoices, "None", "Material");
                RegisterStringChooser(GlobalOptMaterial);
                GlobalOptMaterial.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpPopup = CreateScrollablePopup(GlobalOptMaterial, true);
                tmpPopup.labelWidth = 160f;
                globalColorsOptUIs.Add((UIDynamic) tmpPopup);
                
                GlobalOptMatDetSSX = new JSONStorableFloat("Global_Opt_MatDetSSX", 0.05f, -2f, 2f);
                tmpSlider = CreateSlider(GlobalOptMatDetSSX, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Scroll Speed X";
                RegisterFloat(GlobalOptMatDetSSX);
                GlobalOptMatDetSSX.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatDetSSY = new JSONStorableFloat("Global_Opt_MatDetSSY", 0.05f, -2f, 2f);
                tmpSlider = CreateSlider(GlobalOptMatDetSSY, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Scroll Speed Y";
                RegisterFloat(GlobalOptMatDetSSY);
                GlobalOptMatDetSSY.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatSubDetOpacity = new JSONStorableFloat("Global_Opt_MatSubDetOpacity", 0f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptMatSubDetOpacity, true);
                tmpSlider.valueFormat = "F1";
                tmpSlider.label = "Sub-details Opacity";
                RegisterFloat(GlobalOptMatSubDetOpacity);
                GlobalOptMatSubDetOpacity.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);
                
                GlobalOptMatRefraction = new JSONStorableFloat("Global_Opt_MatRefraction", 0.5f, 0f, 1f);
                tmpSlider = CreateSlider(GlobalOptMatRefraction, true);
                tmpSlider.valueFormat = "F2";
                tmpSlider.label = "Details Refraction";
                RegisterFloat(GlobalOptMatRefraction);
                GlobalOptMatRefraction.setCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                globalColorsOptUIs.Add((UIDynamic) tmpSlider);

                GlobalOptMatDetColor = new JSONStorableColor("Global_Opt_MatDetColor", hsvcDefault);
                GlobalOptMatDetColor.setJSONCallbackFunction += (val) => { OnGlobalOptionsChanged(); };
                tmpColor = CreateColorPicker(GlobalOptMatDetColor, true);
                tmpColor.label = "Details Color";
                RegisterColor(GlobalOptMatDetColor);
                globalColorsOptUIs.Add((UIDynamic) tmpColor);
                
				// Fake action	                
                JSONStorableAction fakeFuncDontUseBelow = new JSONStorableAction("- - - - DON'T Use these functions below ↓ - - - - -", () => {});
                RegisterAction(fakeFuncDontUseBelow);
                
                // ******* ACTIONS CONTROLS **********************

                // Actions choices
                actionsChoices = new List<string>();
                ActionSelector = new JSONStorableStringChooser("Action ID", actionsChoices, "", "Action ID") {isStorable=false,isRestorable=false};
                ActionSelector.setCallbackFunction += (val) => { OnToggleActionId(val); };
                UIDynamicPopup DSelsp = CreateScrollablePopup(ActionSelector, true);
                DSelsp.labelWidth = 250f;
                globalControlsUIs.Add((UIDynamic) DSelsp);
                
                UIDynamicButton newActionBtn = CreateButton("New action", true);
                newActionBtn.button.onClick.AddListener( () => { OnAddNewAction(); } );
                setButtonColor(newActionBtn, new Color(0.3f, 0.6f, 0.3f, 1f));
                setButtonTextColor(newActionBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) newActionBtn);
                setupButtonWithLayout(newActionBtn, 250f);
                
                UIDynamicButton deleteActionBtn = CreateButton("Delete action", true);
                deleteActionBtn.button.onClick.AddListener( () => { OnDeleteCurrentAction(); } );
                setButtonColor(deleteActionBtn, new Color(0.6f, 0.3f, 0.3f, 1f));
                setButtonTextColor(deleteActionBtn, new Color(1f, 1f, 1f, 1f));
                globalControlsUIs.Add((UIDynamic) deleteActionBtn);
                ignoreLayoutUI.Add((UIDynamic) deleteActionBtn);
                setupButtonWithoutLayout(deleteActionBtn, 250f, new Vector2(270, -430));
                
                tmpTextfield = createStaticDescriptionText("Action copy Title","<color=#000><size=27><b>COPY ACTION</b></size></color>",true,35,TextAnchor.MiddleRight);
                globalControlsUIs.Add((UIDynamic) tmpTextfield);
                
                CopySelector = new JSONStorableStringChooser("Action Source", copyChoices, "", "Source/Target") {isStorable=false,isRestorable=false};
                UIDynamicPopup DSosp = CreateScrollablePopup(CopySelector, true);
                DSosp.labelWidth = 220f;
                globalControlsUIs.Add((UIDynamic) DSosp);
                
                UIDynamicButton copyActionBtn = CreateButton("Copy", true);
                copyActionBtn.button.onClick.AddListener(() => { OnCopyAction(); });
                setButtonColor(copyActionBtn, new Color(0.8f, 0.8f, 0.1f, 1f));
                globalControlsUIs.Add((UIDynamic) copyActionBtn);
                setupButtonWithLayout(copyActionBtn, 250f);
                
                UIDynamicButton transferActionBtn = CreateButton("Transfer", true);
                transferActionBtn.button.onClick.AddListener(() => { OnTransferAction(); });
                setButtonColor(transferActionBtn, new Color(0.8f, 0.5f, 0f, 1f));
                globalControlsUIs.Add((UIDynamic) transferActionBtn);
                ignoreLayoutUI.Add((UIDynamic) transferActionBtn);
                setupButtonWithoutLayout(transferActionBtn, 250f, new Vector2(270, -660));
                
                // ******* HELP **********************
                helpText = new JSONStorableString("Help",
	                "<size=30><color=#333>" +
	                "This plugin helps you create action panels for your scenes.\n\n" +
	                "<b>Formatting tips:</b>\n" +
	                "You can format the text of actions by using rich text snippets.\n" +
	                "<size=27>" +
	                "- <b>Bold:</b> <<i></i>b>Text</b<i></i>>\n" +
	                "- <b>Italic:</b> <<i></i>i>Text</i<i></i>>\n" +
	                "- <b>Size:</b> <<i></i>size=20>Text</size<i></i>>\n" +
	                "- <b>Color:</b> <<i></i>color=green>Text</color<i></i>>\n" +
	                "- <b>Color:</b> <<i></i>color=#00ffffff>Text</color<i></i>>\n" +
	                "</size>\n\n" +
	                "<b>PLUGIN HELP:</b>\n" +
	                "<b>VAMStoryActions</b> allows you to create <b>walls of buttons</b> in a couple of clicks.\n\n" +
	                "<b>ACTIONS CONTROLS</b>\n" +
	                "This is where you navigate through your actions, as simple as that.\n\n" +
	                "<b>COPY/TRANSFER TRIGGER</b>\n" +
	                "This is the way to copy or transfer an action.\n\n" +
	                "Select an action in the Source/Target dropdown.\n\n" +
	                "Then click either <b>Copy</b> or <b>Transfer</b>.\n\n" +
	                "<b>Copy</b> will grab the values of the selected action in the dropdown and copy its content to the current action. <b>Transfer</b> will send the current action values to the selected action in the dropdown.\n\n" +
	                "<b>GLOBAL OPTIONS</b>\n" +
	                "This is where you can enable a trigger for the action panel.\n\n" +
	                "Click <b>Edit slider trigger</b> to set the actions bound to the slider.\n\n" +
	                "You can also configure here several appearance or behavior options for the panel ( if the UI rotates, its size, alignment etc... ).\n\n" +
	                "<b>BACKUPS</b>\n" +
	                "<b>VAMStoryActions</b> performs an automatic backup every 20 minutes only in edit mode.\n\n" +
	                "You can also restore a backup or save a manual backup here.\n\n" +
	                "You can use this as a <b>preset system</b> if you want..\n\n" +
	                "<b>COLORS/MATERIAL OPTIONS</b>\n" +
	                "This is where you can customize the colors and material appearance of the buttons.\n\n" +
	                "</color></size>"
                );
				
                tmpTextfield = CreateTextField(helpText, true);
                tmpTextfield.height = 850.0f;
                helpUIs.Add((UIDynamic) tmpTextfield);
                
                // Updating UI
                OnChangeRightUISelection();
                
                // Loading the assetbundle containing the fluids
                Request request = new AssetLoader.AssetBundleFromFileRequest {path = PLUGIN_PATH + UI_BUNDLE_PATH, callback = OnBundleLoaded};
                AssetLoader.QueueLoadAssetBundleFromFile(request);
                
                // Renaming callback
                SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
            }
            catch(Exception e)
            {
                SuperController.LogError("VAMStory - Exception caught: " + e);
            }
		}
		
		void Start()
		{
			
		}
		void Update() {
			// Base initialization after assetbundle load
			if (isInit == false && !SuperController.singleton.isLoading && isBundleLoaded == true)
			{
				isInit = true;
				try
				{
					// Everything is loaded, initialize the global options
					OnGlobalOptionsChanged();
					
					// These actions are ready we can broadcast VAMStories to update their selectors
					SendBroadcastToVAMStories();
				}
				catch(Exception e)
				{
					SuperController.LogError("VAMStory - Exception caught: " + e);
				}
			}
			
			// UI Update
			UpdateUIPosition();
		}

		// **************************
		// Functions
		// **************************
		private void InitComponents()
		{
			if (VAMSAUIGO != null)
			{
				// Transforms init
				TRStoryActionsUI = getChildNamed(VAMSAUIGO.transform,"VAMStoryActionsUI");
				TRAudio = getChildNamed(VAMSAUIGO.transform,"VAMStoryActionsAudio");
				TRActionsMainPanel = getChildNamed(VAMSAUIGO.transform,"ActionsMainPanel");
				TRActionsPanel = getChildNamed(VAMSAUIGO.transform,"ActionsPanel");
				TRActionBtnModel = getChildNamed(VAMSAUIGO.transform,"ActionBtnModel");
				TRActionSlider = getChildNamed(VAMSAUIGO.transform,"ActionSlider");
				TRActionSliderBg = getChildNamed(TRActionSlider,"Background");
				TRActionSliderFill = getChildNamed(TRActionSlider,"Fill");
				TRHideUIGrp = getChildNamed(VAMSAUIGO.transform,"HideGroup");
				TRHideUIBtn = getChildNamed(VAMSAUIGO.transform,"HideBtn");
				TRShowUIGrp = getChildNamed(VAMSAUIGO.transform,"ShowGroup");
				TRShowUIBtn = getChildNamed(VAMSAUIGO.transform,"ShowBtn");
				
				// Components init
				CANStoryActionsUI = TRStoryActionsUI.GetComponent<Canvas>();
				ASAudio = TRAudio.GetComponent<AudioSource>();
				CGActionsPanel = TRActionsPanel.GetComponent<CanvasGroup>();
				GLGActionsPanel = TRActionsPanel.GetComponent<GridLayoutGroup>();
				LEActionsPanel = TRActionsPanel.GetComponent<LayoutElement>();
				CGActionSlider = TRActionSlider.GetComponent<CanvasGroup>();
				SLActionSlider = TRActionSlider.GetComponent<Slider>();
				IMGActionSliderBg = TRActionSliderBg.GetComponent<Image>();
				IMGActionSliderFill = TRActionSliderFill.GetComponent<Image>();
				BTNHideUI = TRHideUIBtn.GetComponent<Button>();
				IMGHideUI = TRHideUIBtn.GetComponent<RawImage>();
				BTNShowUI = TRShowUIBtn.GetComponent<Button>();
				IMGShowUI = TRShowUIBtn.GetComponent<RawImage>();
				
				// Disabling the default action model just in case
				TRActionBtnModel.gameObject.SetActive(false);
				
				// Enabling or disabling the slider depending on the global options
				TRActionSlider.gameObject.SetActive(GlobalOptEnableSlider.val);
				
				// Adding Canvas to VAM's controller
				TRStoryActionsUI.gameObject.AddComponent<IgnoreCanvas>();
				SuperController.singleton.AddCanvas(CANStoryActionsUI);

				// Hiding toggle on UI
				TRShowUIGrp.gameObject.SetActive(false);
				
				// Slider action
				SLActionSlider.onValueChanged.AddListener(delegate { OnChangeSliderValue(); } );
				
				// Hide/Show buttons actions
				BTNHideUI.onClick.AddListener(delegate { OnClickHideUIButton(); } );
				BTNShowUI.onClick.AddListener(delegate { OnClickShowUIButton(); } );
			}
		}
		
		private void CreateDefaultAction()
		{
			StoryAction newAction = new StoryAction(this, 0);
			storyActions.Add(newAction);
			RefreshSelectors(0);
		}

		private GameObject CreateActionComponents()
		{
			GameObject newActionGO = Instantiate(TRActionBtnModel.gameObject);
			newActionGO.transform.SetParent(TRActionsPanel, false);
			newActionGO.SetActive(true);

			return newActionGO;
		}

		private void LoadActions()
		{
			// The save system is pretty "awkward" as there are not a lot of possibilities to properly create a json file
			// that would help to restore properly the storables natively
			// So
			int actionsCount = 0;
			foreach (string paramKey in actionsDatas.Keys)
			{
				if (paramKey == "id") continue;// Skipping plugin id
				string[] paramSplit = paramKey.Split('-');

				if (paramSplit.Length != 2) continue; // Skipping anything that has not at least 2 entries on its param name
				
				// Only keeping "texts" (it's enough to know what I need to create the UI)
				if (paramSplit[1] == "Text")
				{
					string[] idsSplit = paramSplit[0].Split('_');
					if (idsSplit.Length == 2)
					{
						int splitId = int.Parse(idsSplit[1]);
						if (actionsCount < splitId) actionsCount = splitId;
					}
				}
			}

			// If we did find actions in the scene
			if (actionsCount > 0)
			{
				for (var i = 0; i <= actionsCount; i++)
				{
					StoryAction newAction = new StoryAction(this, i);
					storyActions.Add(newAction);
				}
				
				RefreshSelectors(0);
			}
			// If it's not the case, let's create the default action
			else
			{
				CreateDefaultAction();
			}
		}
		
		protected void ClearAllActionUIs()
		{
			foreach (var ActionItem in storyActions)
			{
				ActionItem.ToggleActionUI(false);			
			}
		}

		protected void RefreshSelectors( int actionId )
		{
			actionsChoices = new List<string>();
			copyChoices = new List<string>();
			copyChoices.Add(""); // The empty default one
			for (var i = 0; i < storyActions.Count; i++)
			{
				actionsChoices.Add("" + i);
				copyChoices.Add("" + i);
			}

			// Updating action selector
			ActionSelector.valNoCallback = "";
			ActionSelector.choices = actionsChoices;
			ActionSelector.val = "" + actionId;

			CopySelector.choices = copyChoices;
		}
		
		protected void ToggleActionUI(bool enabled, int actionId)
		{
			if (storyActions.ElementAtOrDefault(actionId) == null) return;
			storyActions[actionId].ToggleActionUI(enabled);
		}
		
		protected void ToggleUIDynamic(bool enabled, UIDynamic target)
		{
			if (target == null) return;
			
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();

			if (tmpLE != null)
			{
				if (enabled == true)
				{

					tmpLE.transform.localScale = new Vector3(1, 1, 1);
					if( !ignoreLayoutUI.Contains(target) ) tmpLE.ignoreLayout = false;
				}
				else
				{
					tmpLE.transform.localScale = new Vector3(0, 0, 0);
					tmpLE.ignoreLayout = true;
				}
			}
		}
		
		private void setupTextField( UIDynamicTextField target, float fieldHeight, bool disableBackground = true, bool disableScroll = true ) {
			if( disableBackground ) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement tfLayout = target.GetComponent<LayoutElement>();
			tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
			target.height = fieldHeight;
			if( disableScroll ) disableScrollOnText(target);
		}

		private void setupSpacer(UIDynamic target, float newHeight)
		{
			LayoutElement tmpLE = target.GetComponent<LayoutElement>();
			tmpLE.minHeight = newHeight;
			tmpLE.preferredHeight = newHeight;
		}

		private void setupButtonWithLayout(UIDynamicButton target, float newSize)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
		}
		
		private void setupButtonWithoutLayout(UIDynamicButton target, float newSize, Vector2 newPosition)
		{
			ContentSizeFitter tmpCSF = target.button.transform.gameObject.AddComponent<ContentSizeFitter>();
			tmpCSF.horizontalFit = UnityEngine.UI.ContentSizeFitter.FitMode.MinSize;
			LayoutElement tmpLE = target.button.transform.gameObject.GetComponent<LayoutElement>();
			RectTransform tmpRectT = target.button.transform.GetComponent<RectTransform>();
			Rect tmpRect = tmpRectT.rect;
			tmpRectT.pivot = new Vector2(0f, 0.5f);
			tmpLE.minWidth = newSize;
			tmpLE.preferredWidth = newSize;
			tmpLE.ignoreLayout = true;
			tmpRectT.anchoredPosition = newPosition;
		}

		private void setButtonColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.buttonColor = newColor;
		}
		
		private void setButtonTextColor(UIDynamicButton targetBtn, Color newColor)
		{
			targetBtn.textColor = newColor;
		}
		
		public UIDynamicTextField createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight, TextAnchor textAlignment = TextAnchor.UpperLeft ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText) {isStorable=false,isRestorable=false};
			staticDescString.hidden = true;
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			staticDescStringField.UItext.alignment = textAlignment;
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescStringField;
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private Transform getChildNamed( Transform parent, string childName ) {
			foreach( Transform child in parent )
			{
				if( child.name == childName ) {
					return child;
				} else {
					Transform childSearch = getChildNamed(child,childName);
					if (childSearch != null) return childSearch;
				}
			}
        
			return null;
		}

		protected void ToggleActionsUI(bool enabled)
		{
			if (TRStoryActionsUI != null)
			{
				TRStoryActionsUI.gameObject.SetActive(enabled);
			}
		}

		protected void UpdateUIPosition() {
			if (isInit == false || VAMSAUIGO == null) return;
			// UI Rotation
			if (GlobalOptUIRotationDisabled.val != true)
			{
				Vector3 lookPos = VAMSCamera.transform.position - VAMSAUIGO.transform.position;
				Quaternion lookRot = Quaternion.LookRotation(lookPos, Vector3.up);
				float eulerX = 0;
				if (GlobalOptUIConstraintToY.val == false)
				{
					eulerX = lookRot.eulerAngles.x;
				}

				float eulerY = lookRot.eulerAngles.y;
				Quaternion rotation = Quaternion.Euler(eulerX, eulerY, 0);
				VAMSAUIGO.transform.rotation = rotation;
			}
		}
		
		// Saving the datas
		// saveType: 0 = auto, 1 = critical, 2 = manual
		protected void SaveBackup( int saveType = 0 )
		{
			if (SuperController.singleton.gameMode != SuperController.GameMode.Edit) return;
			string fileName = "vamsa-" + GetCleanFilenameFromString(containingAtom.name);
			if (saveType == 0)
			{
				logDebug("VAMStoryActions: Saving automatic backup");
				string backupPath = CleanAndGetBackupPath( fileName, "auto" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 1)
			{
				string backupPath = CleanAndGetBackupPath( fileName, "critical" );
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
			} else if (saveType == 2)
			{
				string backupPath = BACKUP_PATH + "\\" + fileName + "-manual-" + DateTime.Now.ToString("yyyyMMdd-HHmmss") + ".json";
				SuperController.singleton.SaveJSON( this.GetJSON(), backupPath);
				logDebug("VAMStoryActions: Manual backup saved in:" + backupPath);
			}
		}

		protected string CleanAndGetBackupPath( string backupName, string backupSuffix )
		{
			// I'm keeping 5 files
			for (int i = 5; i > 0; i--)
			{
				string backupPath = BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json";
				if (FileManagerSecure.FileExists(backupPath) == true && i == 5)
				{
					FileManagerSecure.DeleteFile(backupPath);
				}
				else if(FileManagerSecure.FileExists(backupPath) == true)
				{
					FileManagerSecure.MoveFile(BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + i + ".json", BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-" + ( i + 1 ) + ".json",false);
				}
			}

			return BACKUP_PATH + "\\" + backupName + "-" + backupSuffix + "-1.json";
		}
		
		// Load backup
		protected void LoadBackup()
		{
			SuperController.singleton.fileBrowserUI.defaultPath = BACKUP_PATH;
			SuperController.singleton.fileBrowserUI.SetTitle("Select a backup");
			SuperController.singleton.fileBrowserUI.fileFormat = "json";
			SuperController.singleton.fileBrowserUI.showFiles = true;
			SuperController.singleton.fileBrowserUI.showDirs = true;
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);
			SuperController.singleton.fileBrowserUI.hideExtension = false;
			SuperController.singleton.fileBrowserUI.showInstallFolderInDirectoryList = false;
			List<ShortCut> shortCutList = new List<ShortCut>();

			ShortCut shortCutCustom = new ShortCut();
			shortCutCustom.path = BACKUP_PATH;
			shortCutCustom.displayName = "VAMStoryActions Backups";
			shortCutCustom.flatten = false;
			shortCutList.Add(shortCutCustom);

			SuperController.singleton.fileBrowserUI.shortCuts = shortCutList;			
			SuperController.singleton.fileBrowserUI.Show((val) =>
			{
				if (val != "")
				{
					RestoreBackup(val);
				}
			});
		}

		protected void RestoreBackup(string backupPath)
		{
			// Loading the file
			string backupDatas = SuperController.singleton.ReadFileIntoString(backupPath);
			JSONClass backupJSON = JSON.Parse(backupDatas).AsObject;
			
			// Checking if the save is meant for VAMStory
			if ( !backupJSON.HasKey("id") || !backupJSON["id"].Value.Contains("_VAMStoryActionPlugin.VAMStoryAction") )
			{
				logError("VAMStoryActions: Backup not compatible with VAMStoryActions.");
				return;
			}
			
			logDebug("VAMStoryActions: Restoring " + backupPath);
			
			// Clearing all groups
			foreach (StoryAction act in storyActions)
			{
				act.Delete();
			}

			storyActions.Clear();
			
			// Restoring datas
			this.RestoreFromJSON(backupJSON);
			this.LateRestoreFromJSON(backupJSON);
			
			// Loading Dialogs
			LoadActions();
			
			// Refreshing UI
			RefreshSelectors(0);
		}
		
		// **************************************
		// Callbacks & similar ( internal )
		// **************************************
		private void OnBundleLoaded(Request aRequest) {
			try
			{
				// Loading triggers assets
				SimpleTriggerHandler.LoadAssets();
				
				// Reading my UI bundle
				bundleRequest = aRequest;
				GameObject tmp = bundleRequest.assetBundle.LoadAsset<GameObject>("Assets/VAMStory/prefabs/VAMStoryActions.prefab");
				VAMSAUIGO = Instantiate(tmp);
				
				// Parenting to my current Empty
				VAMSAUIGO.transform.SetParent(containingAtom.reParentObject.GetChild(0).transform, false);
				VAMSAUIGO.transform.localPosition = new Vector3(0, 0, 0);
				VAMSAUIGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
				
				// Grabbing all global assets
				backgroundsChoices = new List<string>();
				backgroundsChoices.Add("Default");
				fontsChoices = new List<string>();
				fontsChoices.Add("arial");
				soundChoices = new List<string>();
				materialChoices = new List<string>();
				materialChoices.Add("None");
				
				// Default assets
				fontsFiles.Add(Resources.GetBuiltinResource(typeof(Font), "Arial.ttf") as Font);
				materialList.Add( bundleRequest.assetBundle.LoadAsset<Material>("Assets/VAMStory/materials/M_VAMS_UI_default.mat") );
				
				// Other assets
				foreach (string bundleFileItem in bundleRequest.assetBundle.GetAllAssetNames())
				{
					if (bundleFileItem.Contains(BACKGROUNDS_PATH))
					{
						backgroundsChoices.Add(bundleFileItem.Replace(BACKGROUNDS_PATH, "").Replace(".png", ""));
						backgroundsSprites.Add( bundleRequest.assetBundle.LoadAsset<Sprite>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(FONTS_PATH))
					{
						fontsChoices.Add(bundleFileItem.Replace(FONTS_PATH, "").Replace(".ttf", "").Replace(".otf", ""));
						fontsFiles.Add( bundleRequest.assetBundle.LoadAsset<Font>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(SOUNDS_PATH))
					{
						soundChoices.Add(bundleFileItem.Replace(SOUNDS_PATH, "").Replace(".ogg", ""));
						soundClips.Add( bundleRequest.assetBundle.LoadAsset<AudioClip>(bundleFileItem) );
					}
					
					if (bundleFileItem.Contains(MATERIALS_PATH))
					{
						materialChoices.Add(bundleFileItem.Replace(MATERIALS_PATH, "").Replace(".mat", ""));
						materialList.Add( bundleRequest.assetBundle.LoadAsset<Material>(bundleFileItem) );
					}
				}
				
				// Settings global storable
				GlobalOptFont.choices = fontsChoices;
				GlobalOptFont.valNoCallback = fontsChoices.First();
				GlobalOptBg.choices = backgroundsChoices;
				GlobalOptBg.valNoCallback = backgroundsChoices.First();
				GlobalOptSoundOnClick.choices = soundChoices;
				GlobalOptSoundOnClick.valNoCallback = soundChoices.First();
				GlobalOptMaterial.choices = materialChoices;
				GlobalOptMaterial.valNoCallback = materialChoices.First();
				
				// Since everything happens after loading the assets
				// I need to restore some of the global options that I just updated
				if (actionsDatas != null)
				{
					GlobalOptFont.RestoreFromJSON( actionsDatas );
					GlobalOptBg.RestoreFromJSON( actionsDatas );
					GlobalOptSoundOnClick.RestoreFromJSON( actionsDatas );
					GlobalOptMaterial.RestoreFromJSON( actionsDatas );
				}
				
				// Initializing all UI components
				InitComponents();
					
				// If there is no data saved, let's create a default action
				if (actionsDatas == null)
				{
					CreateDefaultAction();
				}
				else
				{
					LoadActions();
				}
				
				isBundleLoaded = true;
			}
			catch(Exception e)
			{
				SuperController.LogError("VAMStory - Exception caught: " + e);
			}
		}
		
		protected void OnChangeRightUISelection()
		{
			if (RightUISelector.val == "Actions controls")
			{
				RightUIMainTitle.text = "<color=#000><size=35><b>ACTIONS CONTROLS</b></size></color>";
				
				foreach( UIDynamic uidElem in globalControlsUIs )
				{
					ToggleUIDynamic(true, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalOptionsUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalColorsOptUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}

				foreach( UIDynamic uidElem in helpUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
			} else if (RightUISelector.val == "Global options")
			{
				foreach( UIDynamic uidElem in globalControlsUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalOptionsUIs )
				{
					ToggleUIDynamic(true, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalColorsOptUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
				
				foreach( UIDynamic uidElem in helpUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}

				RightUIMainTitle.text = "<color=#000><size=35><b>GLOBAL OPTIONS</b></size></color>";
			} else if (RightUISelector.val == "Colors/Material options")
			{
				foreach( UIDynamic uidElem in globalControlsUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalOptionsUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalColorsOptUIs )
				{
					ToggleUIDynamic(true, uidElem);
				}
				
				foreach( UIDynamic uidElem in helpUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}

				RightUIMainTitle.text = "<color=#000><size=32><b>COLORS/MATERIAL OPTIONS</b></size></color>";
			} else
			{
				foreach( UIDynamic uidElem in globalControlsUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalOptionsUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}
				
				foreach( UIDynamic uidElem in globalColorsOptUIs )
				{
					ToggleUIDynamic(false, uidElem);
				}

				foreach( UIDynamic uidElem in helpUIs )
				{
					ToggleUIDynamic(true, uidElem);
				}
				
				RightUIMainTitle.text = "<color=#000><size=35><b>HELP</b></size></color>";
			}
		}
		
		protected void OnGlobalOptionsChanged()
		{
			if ( isInit == false ) return;
			
			// Hiding UI toggle
			TRHideUIGrp.gameObject.SetActive(!GlobalOptUIHideToggle.val);
			TRShowUIGrp.gameObject.SetActive(false);
			
			// Enabling/disabling slider
			TRActionSlider.gameObject.SetActive(GlobalOptEnableSlider.val);
			// Triggering the change if it is enabled
			if (GlobalOptEnableSlider.val == true)
			{
				OnChangeSliderValue();
			}
			
			// Changing UI size
			VAMSAUIGO.transform.localScale = new Vector3(GlobalOptUISize.val, GlobalOptUISize.val, GlobalOptUISize.val);
		
			// Changing rotation
			VAMSAUIGO.transform.localRotation = Quaternion.Euler(0f,0f,0f);
			
			// Changing buttons appearance
			foreach (StoryAction act in storyActions)
			{
				act.UpdateSceneComponents();
			}
			
			// UI Alignment
			GLGActionsPanel.childAlignment = GlobalOptUIAlignment.val == "Left" ? TextAnchor.UpperLeft : TextAnchor.UpperRight;
			GLGActionsPanel.startCorner = GlobalOptUIAlignment.val == "Left" ? GridLayoutGroup.Corner.UpperLeft : GridLayoutGroup.Corner.UpperRight;
			
			// Buttons Size / Padding
			GLGActionsPanel.cellSize = new Vector2(GlobalOptBtnWidth.val, GlobalOptBtnHeight.val);
			GLGActionsPanel.spacing = new Vector2(GlobalOptBtnSpacing.val, GlobalOptBtnSpacing.val);
			GLGActionsPanel.padding = new RectOffset(0,0,(int)Mathf.Round(GlobalOptBtnSpacing.val),10);
			LEActionsPanel.minWidth = Mathf.Round(GlobalOptBtnAreaWidth.val);
			LEActionsPanel.preferredWidth = Mathf.Round(GlobalOptBtnAreaWidth.val);
			
			// Main Colors
			Color mainNewCol = new Color( GlobalOptBgColor.colorPicker.currentColor.r, GlobalOptBgColor.colorPicker.currentColor.g, GlobalOptBgColor.colorPicker.currentColor.b, Mathf.Clamp(GlobalOptBgOpacity.val,0.4f,1f) );
			Color sliderHandleNewCol = new Color( GlobalOptTxtColor.colorPicker.currentColor.r, GlobalOptTxtColor.colorPicker.currentColor.g, GlobalOptTxtColor.colorPicker.currentColor.b, Mathf.Clamp(GlobalOptBgOpacity.val,0.6f,1f) );
			
			// Slider
			IMGActionSliderBg.color = mainNewCol;
			IMGActionSliderFill.color = mainNewCol;
			
			// Hide/Show
			IMGHideUI.color = mainNewCol;
			IMGShowUI.color = mainNewCol;
		}
		
		protected void SetMaterialOptions( Image targetImg, Color newCol )
		{
			targetImg.material.SetColor("_DetailsColor",newCol);
			targetImg.material.SetFloat("_DetailsSpeedX", GlobalOptMatDetSSX.val);
			targetImg.material.SetFloat("_DetailsSpeedY", GlobalOptMatDetSSY.val);
			targetImg.material.SetFloat("_SubdetailsOpacity", GlobalOptMatSubDetOpacity.val);
			targetImg.material.SetFloat("_RefractionIntensity", GlobalOptMatRefraction.val);
		}
		
		protected void OnToggleActionId(string actionId)
		{
			ClearAllActionUIs();
			ToggleActionUI(true, int.Parse(ActionSelector.val));
		}
		
		protected void OnAddNewAction()
		{
			int actionId = storyActions.Count;
			StoryAction newAction = new StoryAction(this, actionId);
			storyActions.Add(newAction);
			RefreshSelectors(actionId);
		}

		protected void OnDeleteCurrentAction()
		{
			int actionId = int.Parse(ActionSelector.val);
			
			// We cannot delete the last action
			if (storyActions.Count == 1)
			{
				SuperController.singleton.Alert("<color=green>VAMStoryActions</color> needs to have at least one action.", () => {});
				return;
			}
			
			// Waiting for user confirmation
			SuperController.singleton.Alert("<color=#A90000FF>Deleting action " + actionId + ":\n\n</color><size=30>" + storyActions[actionId].ActionText.val + "</size>\n\nAre you sure?", delegate {OnConfirmActionDeletion();}, () => {});

		}

		protected void OnConfirmActionDeletion()
		{
			// Saving a critical backup
			SaveBackup(1);
			
			// Now delete
			int actionId = int.Parse(ActionSelector.val);
			
			storyActions[actionId].Delete();
			storyActions.RemoveAt(actionId);
			int actCount = 0;
			foreach (StoryAction act in storyActions)
			{
				act.UpdateID(actCount);
				actCount++;
			}
			
			int selectedAction = 0;
			if (storyActions.Count > 1 && actionId > 0)
			{
				selectedAction = actionId - 1;
			}
			
			RefreshSelectors(selectedAction);
		}

		protected void OnCopyAction()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a source action to copy first!", () => {});
				return;
			}
			
			int sourceActionId = int.Parse(CopySelector.val);
			int targetActionId = int.Parse(ActionSelector.val);

			if (sourceActionId == targetActionId)
			{
				SuperController.singleton.Alert("You cannot copy an action on itself!", () => {});
			}
			else
			{
				StoryAction sourceAct = storyActions[sourceActionId];
				StoryAction targetAct = storyActions[targetActionId];

				targetAct.UpdateActionText( sourceAct.ActionText.val );
				targetAct.triggerOnClickAction.RestoreFromJSON(sourceAct.triggerOnClickAction.GetJSON());
			}
		}

		protected void OnTransferAction()
		{
			if (CopySelector.val == "")
			{
				SuperController.singleton.Alert("Select a target action to transfer to first!", () => {});
				return;
			}
			
			int sourceActionId = int.Parse(ActionSelector.val);
			int targetActionId = int.Parse(CopySelector.val);

			if (sourceActionId == targetActionId)
			{
				SuperController.singleton.Alert("You cannot transfer an action on itself!", () => {});
			}
			else
			{
				StoryAction sourceAct = storyActions[sourceActionId];
				StoryAction targetAct = storyActions[targetActionId];

				targetAct.UpdateActionText( sourceAct.ActionText.val );
				targetAct.triggerOnClickAction.RestoreFromJSON(sourceAct.triggerOnClickAction.GetJSON());
				
				SuperController.singleton.Alert("<color=green>VAMStoryActions</color>\n Datas transfered to Action ID " + targetActionId + "!", () => {});
			}
		}
		
		protected void OnActionTextChanged(JSONStorableString target, InputField input)
		{
			target.val = input.text;
		}

		private void PlayOnClickSound()
		{
			ASAudio.PlayOneShot(soundClips[soundChoices.IndexOf(GlobalOptSoundOnClick.val)], GlobalOptSoundOnClickVolume.val);			
		}
		
		private void OnAtomRename(string oldid, string newid)
		{
			GlobalOptSliderTrigger.SyncAtomNames();
			
			foreach( var act in storyActions ) {
				act.triggerOnClickAction.SyncAtomNames();
			}
		}
		
		// **************************************
		// Callbacks & similar ( gameplay )
		// **************************************
		protected void OnSetSliderValue( float sliderVal )
		{
			SLActionSlider.value = sliderVal;
			setSliderValue.valNoCallback = -1f; // To force callback if it is called again
		}
		
		protected void OnChangeSliderValue()
		{
			GlobalOptSliderTrigger.TriggerTransition( SLActionSlider.value );
		}
		
		protected void OnClickHideUIButton()
		{
			// Freeze ? no action
			if (SuperController.singleton.freezeAnimation) return;
			
			PlayOnClickSound();
			
			CGActionsPanel.alpha = 0f;
			CGActionsPanel.blocksRaycasts = false;
			CGActionsPanel.interactable = false;
			
			CGActionSlider.alpha = 0f;
			CGActionSlider.blocksRaycasts = false;
			CGActionSlider.interactable = false;
			
			TRHideUIGrp.gameObject.SetActive(false);
			TRShowUIGrp.gameObject.SetActive(true);
		}
		
		protected void OnClickShowUIButton()
		{
			// Freeze ? no action
			if (SuperController.singleton.freezeAnimation) return;
			
			PlayOnClickSound();
			
			CGActionsPanel.alpha = 1f;
			CGActionsPanel.blocksRaycasts = true;
			CGActionsPanel.interactable = true;
			
			CGActionSlider.alpha = 1f;
			CGActionSlider.blocksRaycasts = true;
			CGActionSlider.interactable = true;
			
			TRHideUIGrp.gameObject.SetActive(true);
			TRShowUIGrp.gameObject.SetActive(false);
		}
		
		// **************************
		// COROUTINES
		// **************************
		public IEnumerator AutoBackup()
		{
			while (true)
			{
				yield return new WaitForSeconds(1200);
				SaveBackup();
			}
		}
		
		// **************************
		// BROADCASTS SEND/RECEIVE
		// **************************
		public void SendBroadcastToVAMStories()
		{
			foreach(Atom at in SuperController.singleton.GetAtoms() ) {
				if( at.type == "Empty" ) {
					foreach (string stId in at.GetStorableIDs())
					{
						if (stId.Contains("_VAMStoryPlugin.VAMStory"))
						{
							at.BroadcastMessage("RefreshVAMStoryActionsSelectors");
						}
					}
				}
			}
		}

		public void ResetFromVAMStory()
		{
			// First, restoring options
			OnGlobalOptionsChanged();
			
			// The resetting the main actions panel dynamically modified by the grid layout group of VAMStory
			RectTransform panelRT = TRActionsMainPanel.gameObject.GetComponent<RectTransform>();
			panelRT.anchoredPosition = new Vector2(0, 0);
			panelRT.anchorMin = new Vector2(0, 0);
			panelRT.anchorMax = new Vector2(1, 1);
			panelRT.offsetMin = new Vector2(0, 335);
			panelRT.offsetMax = new Vector2(0, 0);
			panelRT.sizeDelta = new Vector2(0, -335);

			Rect tmpRect = panelRT.rect;
			tmpRect.x = Mathf.Round(GlobalOptBtnAreaWidth.val) / 2;
			tmpRect.y = -165f;
			tmpRect.width = Mathf.Round(GlobalOptBtnAreaWidth.val);
			tmpRect.height = 165f;
		}

		// **************************
		// OVERRIDES SAVING/LOADING
		// **************************
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			var json = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				json[GlobalOptSliderTrigger.Name] = GlobalOptSliderTrigger.GetJSON(base.subScenePrefix);
				
				foreach( var act in storyActions ) {
					json[act.triggerOnClickAction.Name] = act.triggerOnClickAction.GetJSON(base.subScenePrefix);
				}
			}
			return json;
		}
		
		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
		{
			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
			// Saving the original datas to load all dynamic storables
			actionsDatas = jc;
		}
		
		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				GlobalOptSliderTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}
		
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		private string GetCleanFilenameFromString(string str)
		{
			string tmp = CleanString(str);
			tmp = tmp.ToLower();
			return tmp;
		}
		
		// Thanks: https://stackoverflow.com/questions/11395775/better-way-to-clean-a-string
		private string CleanString(string dirtyString)
		{
			HashSet<char> removeChars = new HashSet<char>(" ?&^$#@!()+-,:;<>’\'-_*\"");
			StringBuilder result = new StringBuilder(dirtyString.Length);
			foreach (char c in dirtyString)
				if (!removeChars.Contains(c)) // prevent dirty chars
					result.Append(c);
			return result.ToString();
		}
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Sync names
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			
			// Stop backup routine
			StopCoroutine("AutoBackup");
			
			// Removing the canvas from VAM's controller
			SuperController.singleton.RemoveCanvas(CANStoryActionsUI);
			
			// I have to clean the transform if it moved to a VAMStory too
			if (TRActionsMainPanel != null)
			{
				Destroy(TRActionsMainPanel.gameObject);
			}

			// Cleaning GameObjects
			Destroy(VAMSAUIGO);
			
			// Cleaning triggers
			GlobalOptSliderTrigger.active = false;
			GlobalOptSliderTrigger.Remove();
			foreach( var act in storyActions ) {
				act.triggerOnClickAction.Remove();
			}

			// Unloading assetbundle
			AssetLoader.DoneWithAssetBundleFromFile(PLUGIN_PATH + UI_BUNDLE_PATH);
		}

		// ***********************************************************
		// EXTERNAL TOOL - Thank you great coders for your content!
		// ***********************************************************

		// *********** Chokapi Decal Maker START *********************
		public Texture2D GetResource(string path)
		{
			if (path == "" || path == null || !FileManagerSecure.FileExists(path) )
			{
				path = GetPackagePath(this) + "Custom/Scripts/VAMStory/default.png";
			}
			Texture2D texture;
			byte[] tmppng = FileManagerSecure.ReadAllBytes(path);
			texture = new Texture2D(1, 1, TextureFormat.RGBA32, false);
			texture.LoadImage(tmppng);
			return texture;
		}
		// *********** Chokapi Decal Maker END *********************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Frequencies Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
	}
}