using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Actions v1.03
//
// Actions for VAMStory

namespace VAMStoryActionPlugin
{
	public partial class VAMStoryAction : MVRScript
    {
	    public class StoryAction
	    {
		    private VAMStoryAction _VAMSA;

		    public int ActionID;

		    public JSONStorableString ActionText;
		    public EventTrigger triggerOnClickAction;

			protected UIDynamicTextField actionTitle;
			protected UIDynamicButton triggerOnClickBtn;
			
			protected InputField ActionTextInputField;
			protected GameObject ActionGO;
			protected Button BTNAction;
			protected Image IMGAction;
			protected Text TXTAction;

		    
		    protected List<UIDynamic> actionsComponentsUI = new List<UIDynamic>();
		    
		    
		    
		    protected JSONClass actionsDatas { get; set; }
		    
		    public StoryAction(VAMStoryAction rootScript, int actionId)
		    {
			    _VAMSA = rootScript;
			    actionsDatas = _VAMSA.actionsDatas;

			    ActionID = actionId;
				    
			    CreateActionComponents();
		    }

		    protected void CreateActionComponents()
		    {
			    // Temporary vars
			    HSVColor hsvcDefault = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
			    UIDynamicTextField tmpTextField;
			    UIDynamicPopup tmpPopup;
			    UIDynamicSlider tmpSlider;
			    UIDynamicColorPicker tmpColor;
			    UIDynamicButton tmpButton;
			    UIDynamicToggle tmpToggle;
			    UIDynamic tmpSpacer;
			    
			    // Creating Action components
				// ******* ACTION TITLE (ID) ***********
				actionTitle = _VAMSA.createStaticDescriptionText("ActionTitle_" + ActionID,"<color=#000><size=35><b>ACTION "+ActionID+"</b></size></color>",false,55,TextAnchor.MiddleLeft);
				actionsComponentsUI.Add((UIDynamic) actionTitle);
				
				// ******* ACTION TEXT ***********
				string newDefaultText = "Type action name...";
				ActionText = new JSONStorableString(SAVE_PREFIX + ActionID + "-Text", "_default_");
				tmpTextField = _VAMSA.CreateTextField(ActionText);
				_VAMSA.setupTextField(tmpTextField, 150f, false, false);
				InputField tmpInp = tmpTextField.UItext.gameObject.AddComponent<InputField>();
				tmpInp.textComponent = tmpTextField.UItext;
				tmpInp.lineType = InputField.LineType.MultiLineNewline;
				actionsComponentsUI.Add((UIDynamic) tmpTextField);
				ActionText.valNoCallback = newDefaultText;
				tmpInp.text = newDefaultText;				
				tmpInp.onValueChanged.AddListener( delegate { _VAMSA.OnActionTextChanged(ActionText,tmpInp); UpdateSceneComponents(); } );
				ActionTextInputField = tmpInp;

				// ******* TRIGGERS ***********
				triggerOnClickAction = new EventTrigger(_VAMSA, SAVE_PREFIX + ActionID + "-TriggerOnClick");
				
				triggerOnClickBtn = _VAMSA.CreateButton("On click trigger", false);
				actionsComponentsUI.Add((UIDynamic)triggerOnClickBtn);
				triggerOnClickBtn.button.onClick.AddListener(() => { triggerOnClickAction.OpenPanelActionStart(); } );
				
				// Registering everything
				RegisterStorables();
				
				// Restoring datas
				if (actionsDatas != null)
				{
					ActionText.RestoreFromJSON( actionsDatas, true, true, false);
					triggerOnClickAction.RestoreFromJSON( actionsDatas, _VAMSA.subScenePrefix, _VAMSA.mergeRestore, true );
				}
				
				// Updating texts inputs
				tmpInp.text = ActionText.val;
				
				// Creating scene components
				ActionGO = _VAMSA.CreateActionComponents();
				
				// Setting name
				ActionGO.name = "action-btn-" + ActionID;
				
				// Settings components vars
				BTNAction = ActionGO.GetComponent<Button>();
				IMGAction = ActionGO.GetComponent<Image>();
				TXTAction = ActionGO.GetComponentInChildren<Text>();
				
				// Setting click trigger
				BTNAction.onClick.AddListener(delegate { DoClickTrigger(); } );
				
				// Updating values
				UpdateSceneComponents();
		    }

		    public void UpdateActionText(string val)
		    {
			    ActionTextInputField.text = val;
		    }
		    
		    public void ToggleActionUI( bool enabled )
		    {
			    foreach( UIDynamic uidElem in actionsComponentsUI )
			    {
				    _VAMSA.ToggleUIDynamic(enabled, uidElem);
			    }
		    }

		    public void DoClickTrigger()
		    {
			    // Freeze ? no action
			    if (SuperController.singleton.freezeAnimation) return;
			    
			    _VAMSA.PlayOnClickSound();
			    triggerOnClickAction.Trigger();
		    }

		    public void UpdateID(int newActionId)
		    {
			    ActionID = newActionId;

			    // Updating action components (param names for the save)
			    UpdateComponentsIDs();
		    }

		    public void UpdateSceneComponents()
		    {
			    if (ActionGO == null) return;
			    
			    // Components
			    TXTAction.text = ActionText.val;
			    
			    // Font size
			    TXTAction.fontSize = (int)_VAMSA.GlobalOptFontSize.val;
			    
			    // Font + color
			    Font tmpFont = _VAMSA.fontsFiles[ _VAMSA.fontsChoices.IndexOf(_VAMSA.GlobalOptFont.val) ];
			    Color fontNewCol = new Color( _VAMSA.GlobalOptTxtColor.colorPicker.currentColor.r, _VAMSA.GlobalOptTxtColor.colorPicker.currentColor.g, _VAMSA.GlobalOptTxtColor.colorPicker.currentColor.b, 1f );
			    TXTAction.font = tmpFont;
			    TXTAction.color = fontNewCol;

			    // Background sprite
			    // Grabbing material first
			    
			    string bgValue = _VAMSA.GlobalOptBg.val;
			    if (bgValue != "Default" && _VAMSA.backgroundsChoices.IndexOf(bgValue) != -1)
			    {
				    IMGAction.sprite = _VAMSA.backgroundsSprites[ _VAMSA.backgroundsChoices.IndexOf(bgValue) - 1 ];
			    }
			    else
			    {
				    IMGAction.sprite = _VAMSA.backgroundsSprites[ _VAMSA.backgroundsChoices.IndexOf("basic_01") - 1 ];
			    }
			    
			    // Material value
			    Material tmpMat = _VAMSA.materialList[_VAMSA.materialChoices.IndexOf(_VAMSA.GlobalOptMaterial.val)];
			    // I have to do that to instantiate materials
			    Material matAction = new Material(tmpMat);
			    // Setting mat
			    IMGAction.material = matAction;
			    // Material setup
			    Color detActionNewCol = new Color( _VAMSA.GlobalOptMatDetColor.colorPicker.currentColor.r, _VAMSA.GlobalOptMatDetColor.colorPicker.currentColor.g, _VAMSA.GlobalOptMatDetColor.colorPicker.currentColor.b, 1f );
			    _VAMSA.SetMaterialOptions(IMGAction,detActionNewCol);

			    
			    // Background color
			    Color bgNewCol = new Color( _VAMSA.GlobalOptBgColor.colorPicker.currentColor.r, _VAMSA.GlobalOptBgColor.colorPicker.currentColor.g, _VAMSA.GlobalOptBgColor.colorPicker.currentColor.b, _VAMSA.GlobalOptBgOpacity.val );
			    IMGAction.color = bgNewCol;
		    }
		    
		    protected void UpdateComponentsIDs()
		    {
			    // Deregister
			    DeregisterStorables();

			    // Update
			    ActionText.name = SAVE_PREFIX + ActionID + "-Text";
			    triggerOnClickAction.Name = SAVE_PREFIX + ActionID + "-TriggerOnClick";
			    actionTitle.text = "<color=#000><size=35><b>ACTION " + ActionID + "</b></size></color>";
			    
			    // Register
			    RegisterStorables();
		    }
		    
			public void Delete() {
				// Clearing gameobject
				Destroy(ActionGO);
				
				// Clearing the UI
				Destroy(ActionTextInputField);
				_VAMSA.RemoveTextField(actionTitle);
				_VAMSA.RemoveTextField(ActionText);
		    	_VAMSA.RemoveButton(triggerOnClickBtn);

				// Unregistering vars
				DeregisterStorables();

				// Removing triggers
				triggerOnClickAction.Remove();
				
				// Clearing data from the loaded data of the scene (to prevent re-restore after deleting)
				// This only happens when you load a scene and have data loaded from the json
				if (actionsDatas != null)
				{
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-Text");
					actionsDatas.Remove(SAVE_PREFIX + ActionID + "-TriggerOnClick");
				}
			}

			protected void DeregisterStorables()
			{
				_VAMSA.DeregisterString(ActionText);
			}
			
			protected void RegisterStorables()
			{
				_VAMSA.RegisterString(ActionText);
			}
	    }
    }
}