using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMStory Director v1.01
//
// Director for VAMStory

namespace VAMStoryDirectorPlugin
{
	public partial class VAMStoryDirector : MVRScript
    {
	    public class StoryTrigger
	    {
		    private VAMStoryDirector _VAMSD;

		    public int TriggerID;

		    public JSONStorableString TriggerName;
		    public EventTrigger triggerAction;
		    public EventTrigger triggerActionDesktop;
		    public EventTrigger triggerActionVR;
		    public JSONStorableFloat TriggerDelay;

			protected UIDynamicTextField triggerTitle;
			protected UIDynamicTextField triggerGlobTitle;
			protected UIDynamicTextField triggerDeskTitle;
			protected UIDynamicTextField triggerVRTitle;
			protected UIDynamicButton triggerEditBtn;
			protected UIDynamicButton triggerDesktopEditBtn;
			protected UIDynamicButton triggerVREditBtn;
			
			protected InputField TriggerNameInputField;
			
		    protected List<UIDynamic> triggersComponentsUI = new List<UIDynamic>();

		    protected JSONClass directorDatas { get; set; }
		    
		    public StoryTrigger(VAMStoryDirector rootScript, int triggerId)
		    {
			    _VAMSD = rootScript;
			    directorDatas = _VAMSD.directorDatas;

			    TriggerID = triggerId;
				    
			    CreateTriggerComponents();
		    }

		    protected void CreateTriggerComponents()
		    {
			    // Temporary vars
			    UIDynamicTextField tmpTextfield;
			    UIDynamicPopup tmpPopup;
			    UIDynamicSlider tmpSlider;
			    UIDynamicColorPicker tmpColor;
			    UIDynamicButton tmpButton;
			    UIDynamicToggle tmpToggle;
			    UIDynamic tmpSpacer;
			    
			    // Creating Trigger components
				// ******* TRIGGER TITLE (ID) ***********
				triggerTitle = _VAMSD.createStaticDescriptionText("TriggerTitle_" + TriggerID,"<color=#000><size=35><b>TRIGGER "+TriggerID+"</b></size></color>",false,55,TextAnchor.MiddleLeft);
				triggersComponentsUI.Add((UIDynamic) triggerTitle);
				
				// ******* TRIGGER TEXT ***********
				TriggerName = new JSONStorableString(SAVE_PREFIX + TriggerID + "-Name", "_default_");
				tmpTextfield = _VAMSD.CreateTextField(TriggerName);
				_VAMSD.setupTextField(tmpTextfield, 150f, false, false);
				InputField tmpInp = tmpTextfield.UItext.gameObject.AddComponent<InputField>();
				tmpInp.textComponent = tmpTextfield.UItext;
				tmpInp.lineType = InputField.LineType.MultiLineNewline;
				triggersComponentsUI.Add((UIDynamic) tmpTextfield);
				TriggerName.valNoCallback = ""+TriggerID;
				tmpInp.text = ""+TriggerID;				
				tmpInp.onValueChanged.AddListener( delegate { OnTriggerNameChanged(); } );
				tmpInp.onEndEdit.AddListener(delegate { OnTriggerNameEdited(); } );
				TriggerNameInputField = tmpInp;

				// ******* TRIGGERS ***********
				triggerAction = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-Trigger");
				triggerActionDesktop = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-TriggerDesktop");
				triggerActionVR = new EventTrigger(_VAMSD, SAVE_PREFIX + TriggerID + "-TriggerVR");
				
				triggerGlobTitle = _VAMSD.createStaticDescriptionText("Backups Title","<color=#000><size=25><b>ALWAYS TRIGGERS</b></size></color>",false,35,TextAnchor.MiddleLeft);
				triggersComponentsUI.Add((UIDynamic) triggerGlobTitle);
				
				triggerEditBtn = _VAMSD.CreateButton("Edit Global trigger", false);
				triggersComponentsUI.Add((UIDynamic)triggerEditBtn);
				triggerEditBtn.button.onClick.AddListener(() => { triggerAction.OpenPanelActionStart(); } );
				
				triggerDeskTitle = _VAMSD.createStaticDescriptionText("Backups Title","<color=#000><size=25><b>TRIGGERS ONLY IN DESKTOP MODE</b></size></color>",false,35,TextAnchor.MiddleLeft);
				triggersComponentsUI.Add((UIDynamic) triggerDeskTitle);
				
				triggerDesktopEditBtn = _VAMSD.CreateButton("Edit Desktop trigger", false);
				triggersComponentsUI.Add((UIDynamic)triggerDesktopEditBtn);
				triggerDesktopEditBtn.button.onClick.AddListener(() => { triggerActionDesktop.OpenPanelActionStart(); } );
				
				triggerVRTitle = _VAMSD.createStaticDescriptionText("Backups Title","<color=#000><size=25><b>TRIGGERS ONLY IN VR MODE</b></size></color>",false,35,TextAnchor.MiddleLeft);
				triggersComponentsUI.Add((UIDynamic) triggerVRTitle);
				
				triggerVREditBtn = _VAMSD.CreateButton("Edit VR trigger", false);
				triggersComponentsUI.Add((UIDynamic)triggerVREditBtn);
				triggerVREditBtn.button.onClick.AddListener(() => { triggerActionVR.OpenPanelActionStart(); } );
				
				// ******* TRIGGER START DELAY ***********
				TriggerDelay = new JSONStorableFloat(SAVE_PREFIX + TriggerID + "-Delay", 0f, 0f, 5f, false);
				tmpSlider = _VAMSD.CreateSlider(TriggerDelay);
				tmpSlider.valueFormat = "F1";
				tmpSlider.label = "Trigger delay";
				triggersComponentsUI.Add((UIDynamic) tmpSlider);
				
				// Registering everything
				RegisterStorables();
				
				// Restoring datas
				if (directorDatas != null)
				{
					TriggerName.RestoreFromJSON( directorDatas, true, true, false);
					triggerAction.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					triggerActionDesktop.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					triggerActionVR.RestoreFromJSON( directorDatas, _VAMSD.subScenePrefix, _VAMSD.mergeRestore, true );
					TriggerDelay.RestoreFromJSON( directorDatas, true, true, false);
				}
				
				// Updating texts inputs
				tmpInp.text = TriggerName.val;
				
				// Updating dict
				UpdateIdsToNameDict();
		    }

		    public void UpdateTriggerName(string val)
		    {
			    TriggerNameInputField.text = val;
		    }
		    
		    public void ToggleTriggerUI( bool enabled )
		    {
			    foreach( UIDynamic uidElem in triggersComponentsUI )
			    {
				    _VAMSD.ToggleUIDynamic(enabled, uidElem);
			    }
		    }

		    public void UpdateID(int newTriggerId)
		    {
			    TriggerID = newTriggerId;

			    // Updating trigger components (param names for the save)
			    UpdateComponentsIDs();
		    }

		    protected void UpdateIdsToNameDict()
		    {
			    if (!_VAMSD.triggersIdsToNames.ContainsKey(TriggerID))
			    {
				    _VAMSD.triggersIdsToNames.Add(TriggerID,TriggerName.val);
			    }
			    else
			    {
				    _VAMSD.triggersIdsToNames[TriggerID] = TriggerName.val;
			    }
		    } 
		    
		    protected void UpdateComponentsIDs()
		    {
			    // Deregister
			    DeregisterStorables();

			    // Update
			    TriggerName.name = SAVE_PREFIX + TriggerID + "-Name";
			    triggerAction.Name = SAVE_PREFIX + TriggerID + "-Trigger";
			    triggerActionDesktop.Name = SAVE_PREFIX + TriggerID + "-TriggerDesktop";
			    triggerActionVR.Name = SAVE_PREFIX + TriggerID + "-TriggerVR";
			    TriggerDelay.name = SAVE_PREFIX + TriggerID + "-Delay";
			    triggerTitle.text = "<color=#000><size=35><b>TRIGGER " + TriggerID + "</b></size></color>";
			    
			    // Register
			    RegisterStorables();
			    
			    // Updating dict
			    UpdateIdsToNameDict();
		    }
		    
			public void Delete() {
				// Clearing the UI
				Destroy(TriggerNameInputField);
				_VAMSD.RemoveTextField(triggerTitle);
				_VAMSD.RemoveTextField(TriggerName);
				_VAMSD.RemoveTextField(triggerGlobTitle);
				_VAMSD.RemoveTextField(triggerDeskTitle);
				_VAMSD.RemoveTextField(triggerVRTitle);
				_VAMSD.RemoveSlider(TriggerDelay);
		    	_VAMSD.RemoveButton(triggerEditBtn);
		    	_VAMSD.RemoveButton(triggerDesktopEditBtn);
		    	_VAMSD.RemoveButton(triggerVREditBtn);

				// Unregistering vars
				DeregisterStorables();

				// Removing triggers
				triggerAction.Remove();
				triggerActionDesktop.Remove();
				triggerActionVR.Remove();
				
				// Clearing data from the loaded data of the scene (to prevent re-restore after deleting)
				// This only happens when you load a scene and have data loaded from the json
				if (directorDatas != null)
				{
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-Name");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-Trigger");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-TriggerDesktop");
					directorDatas.Remove(SAVE_PREFIX + TriggerID + "-TriggerVR");
				}
			}

			protected void DeregisterStorables()
			{
				_VAMSD.DeregisterString(TriggerName);
				_VAMSD.DeregisterFloat(TriggerDelay);
			}
			
			protected void RegisterStorables()
			{
				_VAMSD.RegisterString(TriggerName);
				_VAMSD.RegisterFloat(TriggerDelay);
			}
			
			protected void OnTriggerNameChanged()
			{
				TriggerName.val = TriggerNameInputField.text;
			}
		
			protected void OnTriggerNameEdited()
			{
				string newName = _VAMSD.GetCleanAndUniqueTriggerName(TriggerNameInputField.text, TriggerID);
				TriggerNameInputField.text = newName;
				UpdateIdsToNameDict();
				_VAMSD.RefreshSelectors(_VAMSD.triggersIdsToNames.FirstOrDefault(el => el.Value == newName).Key);
			}
	    }
    }
}